#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Dict, Union

from pythonwrench.warnings import deprecated_alias, deprecated_function
from torch import Tensor
from torchwrench.nn.functional.padding import pad_dim
from torchwrench.nn.functional.predicate import is_stackable
from torchwrench.utils.data.collate import AdvancedCollateDict, CollateDict, KeyMode


class BasicCollate(CollateDict):
    """Collate object for :class:`~torch.utils.data.dataloader.DataLoader`.

    Merge lists in dicts into a single dict of lists. No padding is applied.
    """

    def __init__(self, key_mode: KeyMode = "intersect") -> None:
        super().__init__(key_mode)


class AdvancedCollate(AdvancedCollateDict):
    """Advanced collate object for :class:`~torch.utils.data.dataloader.DataLoader`.

    Merge lists in dicts into a single dict of lists.
    Audio will be padded if a fill value is given in `__init__`.

    .. code-block:: python
        :caption:  Example

        >>> collate = AdvancedCollate({"audio": 0.0})
        >>> loader = DataLoader(..., collate_fn=collate)
        >>> next(iter(loader))
        ... {"audio": tensor([[...]]), ...}

    """

    def __init__(self, fill_values: Dict[str, Union[float, int]]) -> None:
        super().__init__(fill_values, key_mode="intersect")


@deprecated_function()
def pad_last_dim(tensor: Tensor, target_length: int, pad_value: float) -> Tensor:
    """Left padding tensor at last dim.

    :param tensor: Tensor of at least 1 dim. (..., T)
    :param target_length: Target length of the last dim. If target_length <= T, the function has no effect.
    :param pad_value: Fill value used to pad tensor.
    :returns: A tensor of shape (..., target_length).
    """
    return pad_dim(tensor, target_length, pad_value=pad_value)


@deprecated_alias(is_stackable)
def can_be_stacked(*args, **kwargs): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os
import os.path as osp
from pathlib import Path
from typing import Any, Union

logger = logging.getLogger(__name__)


# Public functions
def get_default_root() -> str:
    """Returns the default root directory path.

    If :func:`~aac_datasets.utils.globals.set_default_root` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_DATASETS_ROOT has been set to a string, it will return its value.
    Else it will be equal to "." by default.
    """
    return __get_default_value("root")


def get_default_ffmpeg_path() -> str:
    """Returns the default ffmpeg executable path.

    If :func:`~aac_datasets.utils.globals.set_default_ffmpeg_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_DATASETS_FFMPEG_PATH has been set to a string, it will return its value.
    Else it will be equal to "ffmpeg" by default.
    """
    return __get_default_value("ffmpeg")


def get_default_ytdlp_path() -> str:
    """Returns the default yt-dlp executable path.

    If :func:`~aac_datasets.utils.globals.set_default_ytdlp_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_DATASETS_YTDLP_PATH has been set to a string, it will return its value.
    Else it will be equal to "yt-dlp" by default.
    """
    return __get_default_value("ytdlp")


def get_default_zip_path() -> str:
    """Returns the default zip executable path.

    If :func:`~aac_datasets.utils.globals.set_default_zip_path` has been used before with a string argument, it will return the value given to this function.
    Else if the environment variable AAC_DATASETS_ZIP_PATH has been set to a string, it will return its value.
    Else it will be equal to "zip" by default.
    """
    return __get_default_value("zip")


def set_default_root(cache_path: Union[str, Path, None]) -> None:
    """Override default root directory path."""
    __set_default_value("root", cache_path)


def set_default_ffmpeg_path(tmp_path: Union[str, Path, None]) -> None:
    """Override default ffmpeg executable path."""
    __set_default_value("ffmpeg", tmp_path)


def set_default_ytdlp_path(java_path: Union[str, Path, None]) -> None:
    """Override default yt-dl executable path."""
    __set_default_value("ytdlp", java_path)


def set_default_zip_path(tmp_path: Union[str, Path, None]) -> None:
    """Override default zip executable path."""
    __set_default_value("zip", tmp_path)


# Private functions
def _get_root(root: Union[str, Path, None] = None) -> str:
    return __get_value("root", root)


def _get_ffmpeg_path(ffmpeg_path: Union[str, Path, None] = None) -> str:
    return __get_value("ffmpeg", ffmpeg_path)


def _get_ytdlp_path(ytdlp_path: Union[str, Path, None] = None) -> str:
    return __get_value("ytdlp", ytdlp_path)


def _get_zip_path(zip_path: Union[str, Path, None] = None) -> str:
    return __get_value("zip", zip_path)


def __get_default_value(value_name: str) -> str:
    values = __DEFAULT_GLOBALS[value_name]["values"]
    process_func = __DEFAULT_GLOBALS[value_name]["process"]

    for source, value_or_env_varname in values.items():
        if source.startswith("env"):
            value = os.getenv(value_or_env_varname, None)
        else:
            value = value_or_env_varname

        if value is not None:
            value = process_func(value)
            return value

    logger.error(f"Values: {values}")
    raise RuntimeError(
        f"Invalid default value for {value_name=}. (all default values are None)"
    )


def __set_default_value(
    value_name: str,
    value: Any,
) -> None:
    __DEFAULT_GLOBALS[value_name]["values"]["user"] = value


def __get_value(value_name: str, value: Any = None) -> Any:
    if value is None or value is ...:
        return __get_default_value(value_name)
    else:
        process_func = __DEFAULT_GLOBALS[value_name]["process"]
        value = process_func(value)
        return value


def __process_path(value: Union[str, Path, None]) -> Union[str, None]:
    if value is None or value is ...:
        return None

    value = str(value)
    value = osp.expanduser(value)
    value = osp.expandvars(value)
    return value


__DEFAULT_GLOBALS = {
    "root": {
        "values": {
            "user": None,
            "env": "AAC_DATASETS_ROOT",
            "package": ".",
        },
        "process": __process_path,
    },
    "ytdlp": {
        "values": {
            "user": None,
            "env": "AAC_DATASETS_YTDLP_PATH",
            "package": "yt-dlp",
        },
        "process": __process_path,
    },
    "ffmpeg": {
        "values": {
            "user": None,
            "env": "AAC_DATASETS_FFMPEG_PATH",
            "package": "ffmpeg",
        },
        "process": __process_path,
    },
    "zip": {
        "values": {
            "user": None,
            "env": "AAC_DATASETS_ZIP_PATH",
            "package": "zip",
        },
        "process": __process_path,
    },
}



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from aac_datasets.datasets import Clotho


class TestLoad(TestCase):
    def test_load_clotho(self) -> None:
        ds = Clotho(subset="val")
        item = ds[0]
        assert isinstance(item, dict)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import unittest
from unittest import TestCase

from aac_datasets.datasets.base import _flat_raw_data, _unflat_raw_data


class TestDatasetBase(TestCase):
    def test_flat_raw_data(self) -> None:
        raw_data = {
            "captions": [["a1", "a2", "a3"], ["b1"], ["c1", "c2"], []],
            "index": list(range(1, 5)),
        }
        expected_flat = {
            "captions": [["a1"], ["a2"], ["a3"], ["b1"], ["c1"], ["c2"], []],
            "index": [1, 1, 1, 2, 3, 3, 4],
        }
        expected_sizes = [3, 1, 2, 0]

        raw_data_flat_out, sizes_out = _flat_raw_data(raw_data, "captions")
        raw_data_out = _unflat_raw_data(raw_data_flat_out, sizes_out)

        assert raw_data_flat_out == expected_flat
        assert sizes_out == expected_sizes
        assert raw_data_out == raw_data


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
from pathlib import Path
from typing import Any, Callable, ClassVar, Dict, Iterable, List, Optional, Union

import torch
import torchaudio
from torch import Tensor
from typing_extensions import NotRequired, TypedDict

from aac_datasets.datasets.base import AACDataset
from aac_datasets.datasets.functional.audiocaps import (
    AudioCapsCard,
    AudioCapsSubset,
    AudioCapsVersion,
    _get_audio_subset_dpath,
    download_audiocaps_dataset,
    load_audiocaps_dataset,
)
from aac_datasets.utils.globals import _get_ffmpeg_path, _get_root, _get_ytdlp_path
from aac_datasets.utils.typing import AudioMetaData

logger = logging.getLogger(__name__)


class AudioCapsItem(TypedDict):
    r"""Class representing a single AudioCaps item."""

    # Common attributes
    audio: Tensor
    captions: List[str]
    dataset: str
    fname: str
    index: int
    subset: AudioCapsSubset
    sr: int
    duration: float
    # AudioCaps-specific attributes
    audiocaps_ids: List[int]
    start_time: int
    tags: NotRequired[List[int]]
    youtube_id: str


class AudioCaps(AACDataset[AudioCapsItem]):
    r"""Unofficial AudioCaps PyTorch dataset.

    Subsets available are 'train', 'val' and 'test'.

    Audio is a waveform tensor of shape (1, n_times) of 10 seconds max, sampled at 32kHz by default.
    Target is a list of strings containing the captions.
    The 'train' subset has only 1 caption per sample and 'val' and 'test' have 5 captions.
    Download from YouTube requires 'yt-dlp' and 'ffmpeg' commands.

    /!\ YouTube website can sometimes block your IP when downloading audio with the error:
        Sign in to confirm you’re not a bot. Use --cookies-from-browser or --cookies for the authentication.
        See https://github.com/yt-dlp/yt-dlp/wiki/FAQ#how-do-i-pass-cookies-to-yt-dlp for how to manually pass cookies.
        Also see https://github.com/yt-dlp/yt-dlp/wiki/Extractors#exporting-youtube-cookies for tips on effectively exporting YouTube cookies.

    You can pass yt-dlp args with `ytdlp_opts` argument, e.g. `AudioCaps(ytdlp_opts=["--cookies-from-browser", "firefox"])`.

    See also: AudioCaps paper : https://www.aclweb.org/anthology/N19-1011.pdf

    .. code-block:: text
        :caption:  Dataset folder tree (for version v1)

        {root}
        └── AUDIOCAPS
            ├── csv_files_v1
            │   ├── train.csv
            │   ├── val.csv
            │   └── test.csv
            └── audio_32000Hz
                ├── train
                │   └── (46231/49838 flac files, ~42G for 32kHz)
                ├── val
                │   └── (465/495 flac files, ~425M for 32kHz)
                └── test
                    └── (913/975 flac files, ~832M for 32kHz)

    """

    # Common globals
    CARD: ClassVar[AudioCapsCard] = AudioCapsCard()

    # Initialization
    def __init__(
        self,
        # Common args
        root: Union[str, Path, None] = None,
        subset: AudioCapsSubset = AudioCapsCard.DEFAULT_SUBSET,
        download: bool = False,
        transform: Optional[Callable[[AudioCapsItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        *,
        # AudioCaps-specific args
        audio_duration: float = 10.0,
        audio_format: str = "flac",
        audio_n_channels: int = 1,
        download_audio: bool = True,
        exclude_removed_audio: bool = True,
        ffmpeg_path: Union[str, Path, None] = None,
        flat_captions: bool = False,
        max_workers: Optional[int] = 1,
        sr: int = 32_000,
        with_tags: bool = False,
        ytdlp_path: Union[str, Path, None] = None,
        ytdlp_opts: Iterable[str] = (),
        version: AudioCapsVersion = AudioCapsCard.DEFAULT_VERSION,
        num_dl_attempts: int = 2,
    ) -> None:
        """
        :param root: Dataset root directory.
            The data will be stored in the 'AUDIOCAPS' subdirectory.
            defaults to ".".
        :param subset: The subset of AudioCaps to use. Can be one of :attr:`~AudioCapsCard.SUBSETS`.
            defaults to "train".
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            defaults to False.
        :param transform: The transform to apply to the global dict item. This transform is applied only in getitem method when argument is an integer.
            defaults to None.
        :param verbose: Verbose level.
            defaults to 0.
        :param force_download: If True, force to re-download file even if they exists on disk.
            defaults to False.
        :param verify_files: If True, check hash value when possible.
            defaults to False.

        :param audio_duration: Extracted duration for each audio file in seconds.
            defaults to 10.0.
        :param audio_format: Audio format and extension name.
            defaults to "flac".
        :param audio_n_channels: Number of channels extracted for each audio file.
            defaults to 1.
        :param download_audio: If True, download audio, metadata and labels files. Otherwise it will only donwload metadata and labels files.
            defaults to True.
        :param exclude_removed_audio: If True, the dataset will exclude from the dataset the audio not downloaded from youtube (i.e. not present on disk).
            If False, invalid audios will return an empty tensor of shape (0,).
            defaults to True.
        :param ffmpeg_path: Path to ffmpeg executable file.
            defaults to "ffmpeg".
        :param flat_captions: If True, map captions to audio instead of audio to caption.
            defaults to True.
        :param max_workers: Number of threads to download audio files in parallel.
            Do not use a value too high to avoid "Too Many Requests" error.
            The value None will use `min(32, os.cpu_count() + 4)` workers, which is the default of ThreadPoolExecutor.
            defaults to 1.
        :param sr: The sample rate used for audio files in the dataset (in Hz).
            Since original YouTube videos are recorded in various settings, this parameter allow to download allow audio files with a specific sample rate.
            defaults to 32000.
        :param verify_files: If True, check all file already downloaded are valid.
            defaults to False.
        :param with_tags: If True, load the tags from AudioSet dataset.
            Note: tags needs to be downloaded with download=True & with_tags=True before being used.
            defaults to False.
        :param ytdlp_path: Path to yt-dlp or ytdlp executable.
            defaults to "yt-dlp".
        :param ytdlp_opts: yt-dlp optional arguments.
            defaults to ().
        :param version: The version of the dataset. Can be one of :attr:`~AudioCapsCard.VERSIONS`.
            defaults to 'v1'.
        :param num_dl_attempts: Number of download attempts.
            defaults to 2.
        """
        if subset not in AudioCapsCard.SUBSETS:
            msg = f"Invalid argument {subset=} for AudioCaps. (expected one of {AudioCapsCard.SUBSETS})"
            raise ValueError(msg)

        root = _get_root(root)
        ytdlp_path = _get_ytdlp_path(ytdlp_path)
        ffmpeg_path = _get_ffmpeg_path(ffmpeg_path)

        if download:
            download_audiocaps_dataset(
                root=root,
                subset=subset,
                force=force_download,
                verbose=verbose,
                verify_files=verify_files,
                audio_duration=audio_duration,
                audio_format=audio_format,
                audio_n_channels=audio_n_channels,
                download_audio=download_audio,
                ffmpeg_path=ffmpeg_path,
                max_workers=max_workers,
                sr=sr,
                with_tags=with_tags,
                ytdlp_path=ytdlp_path,
                ytdlp_opts=ytdlp_opts,
                version=version,
                num_dl_attempts=num_dl_attempts,
            )

        raw_data, index_to_name = load_audiocaps_dataset(
            root=root,
            subset=subset,
            verbose=verbose,
            audio_format=audio_format,
            exclude_removed_audio=exclude_removed_audio,
            sr=sr,
            with_tags=with_tags,
        )
        audio_subset_dpath = _get_audio_subset_dpath(root, subset, sr, version)
        size = len(next(iter(raw_data.values())))
        raw_data["dataset"] = [AudioCapsCard.NAME] * size
        raw_data["subset"] = [subset] * size
        raw_data["fpath"] = [
            osp.join(audio_subset_dpath, fname) for fname in raw_data["fname"]
        ]
        raw_data["index"] = list(range(size))

        column_names = list(AudioCapsItem.__required_keys__) + list(  # type: ignore
            AudioCapsItem.__optional_keys__  # type: ignore
        )
        if not with_tags:
            column_names.remove("tags")

        super().__init__(
            raw_data=raw_data,
            transform=transform,
            column_names=column_names,
            flat_captions=flat_captions,
            sr=sr,
            verbose=verbose,
        )

        # Attributes
        self._root = root
        self._subset: AudioCapsSubset = subset
        self._download = download
        self._exclude_removed_audio = exclude_removed_audio
        self._with_tags = with_tags
        self._version: AudioCapsVersion = version
        self._index_to_name = index_to_name

        self.add_online_columns(
            {
                "audio": AudioCaps._load_audio,
                "audio_metadata": AudioCaps._load_audio_metadata,
                "duration": AudioCaps._load_duration,
                "num_channels": AudioCaps._load_num_channels,
                "num_frames": AudioCaps._load_num_frames,
                "sr": AudioCaps._load_sr,
            }
        )

    # Properties
    @property
    def download(self) -> bool:
        return self._download

    @property
    def exclude_removed_audio(self) -> bool:
        return self._exclude_removed_audio

    @property
    def index_to_name(self) -> Dict[int, str]:
        return self._index_to_name

    @property
    def root(self) -> str:
        return self._root

    @property
    def sr(self) -> int:
        return self._sr  # type: ignore

    @property
    def subset(self) -> AudioCapsSubset:
        return self._subset

    @property
    def version(self) -> AudioCapsVersion:
        return self._version

    @property
    def with_tags(self) -> bool:
        return self._with_tags

    # Magic methods
    def __repr__(self) -> str:
        repr_dic = {
            "subset": self._subset,
            "size": len(self),
            "num_columns": len(self.column_names),
            "with_tags": self._with_tags,
            "exclude_removed_audio": self._exclude_removed_audio,
        }
        repr_str = ", ".join(f"{k}={v}" for k, v in repr_dic.items())
        return f"{AudioCapsCard.PRETTY_NAME}({repr_str})"

    # Private methods
    def _load_audio(self, index: int) -> Tensor:
        if not self._raw_data["is_on_disk"][index]:
            return torch.empty((0,))

        fpath = self.get_item(index, "fpath")
        audio, sr = torchaudio.load(fpath)  # type: ignore

        # Sanity check
        if audio.nelement() == 0:
            msg = f"Invalid number of elements ({audio.nelement()}) for audio '{fpath}'. (expected > 0)"
            raise RuntimeError(msg)

        if self._sr is not None and (self._sr != sr):
            msg = f"Invalid sample rate {sr}Hz for audio '{fpath}'. (expected {self._sr}Hz)"
            raise RuntimeError(msg)
        return audio

    def _load_audio_metadata(self, index: int) -> AudioMetaData:
        if not self._raw_data["is_on_disk"][index]:
            return AudioMetaData(-1, -1, -1, -1, "unknown_encoding")
        fpath = self.get_item(index, "fpath")
        audio_metadata = torchaudio.info(fpath)  # type: ignore
        return audio_metadata



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import json
import logging
import os
import os.path as osp
import subprocess
import zipfile
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union, get_args

import pythonwrench as pw
import tqdm
from huggingface_hub import snapshot_download
from huggingface_hub.constants import HUGGINGFACE_HUB_CACHE
from huggingface_hub.utils.tqdm import (
    are_progress_bars_disabled,
    disable_progress_bars,
    enable_progress_bars,
)
from torchwrench.hub.download import download_file, safe_rmdir
from typing_extensions import Literal, TypedDict

from aac_datasets.datasets.functional.common import DatasetCard, LinkInfo
from aac_datasets.utils.globals import _get_root, _get_zip_path

logger = logging.getLogger(__name__)

WavCapsSource = Literal["AudioSet_SL", "BBC_Sound_Effects", "FreeSound", "SoundBible"]
WavCapsSubset = Literal[
    "audioset",
    "bbc",
    "freesound",
    "soundbible",
    "audioset_no_audiocaps_v1",
    "freesound_no_clotho_v2",
]


class WavCapsCard(DatasetCard):
    ANNOTATIONS_CREATORS: Tuple[str, ...] = ("machine-generated",)
    CAPTIONS_PER_AUDIO: Dict[WavCapsSubset, int] = {
        "audioset": 1,
        "bbc": 1,
        "freesound": 1,
        "soundbible": 1,
        "audioset_no_audiocaps_v1": 1,
        "freesound_no_clotho_v2": 1,
    }
    CITATION: str = r"""
    @article{mei2023WavCaps,
        title        = {Wav{C}aps: A {ChatGPT}-Assisted Weakly-Labelled Audio Captioning Dataset for Audio-Language Multimodal Research},
        author       = {Xinhao Mei and Chutong Meng and Haohe Liu and Qiuqiang Kong and Tom Ko and Chengqi Zhao and Mark D. Plumbley and Yuexian Zou and Wenwu Wang},
        year         = 2023,
        journal      = {arXiv preprint arXiv:2303.17395},
        url          = {https://arxiv.org/pdf/2303.17395.pdf}
    }
    """
    DEFAULT_REVISION: str = "85a0c21e26fa7696a5a74ce54fada99a9b43c6de"
    DEFAULT_SUBSET: WavCapsSubset = "audioset_no_audiocaps_v1"
    DESCRIPTION: str = "WavCaps: A ChatGPT-Assisted Weakly-Labelled Audio Captioning Dataset for Audio-Language Multimodal Research."
    EXPECTED_SIZES: Dict[WavCapsSource, int] = {
        "AudioSet_SL": 108317,
        "BBC_Sound_Effects": 31201,
        "FreeSound": 262300,
        "SoundBible": 1320,  # note: 1232 according to github+hf, but found 1320 => seems that archive contains more data than in json
    }
    HOMEPAGE = "https://huggingface.co/datasets/cvssp/WavCaps"
    LANGUAGE: Tuple[str, ...] = ("en",)
    LANGUAGE_DETAILS: Tuple[str, ...] = ("en-US",)
    NAME: str = "wavcaps"
    PRETTY_NAME: str = "WavCaps"
    REPO_ID: str = "cvssp/WavCaps"
    SOURCES: Tuple[WavCapsSource, ...] = get_args(WavCapsSource)
    SUBSETS: Tuple[WavCapsSubset, ...] = get_args(WavCapsSubset)
    SAMPLE_RATE: int = 32_000  # Hz
    SIZE_CATEGORIES: Tuple[str, ...] = ("100K<n<1M",)
    TASK_CATEGORIES: Tuple[str, ...] = ("audio-to-text", "text-to-audio")


def load_wavcaps_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: WavCapsSubset = WavCapsCard.DEFAULT_SUBSET,
    verbose: int = 0,
    *,
    # WavCaps-specific args
    hf_cache_dir: Optional[str] = None,
    revision: Optional[str] = None,
) -> Dict[str, List[Any]]:
    """Load WavCaps metadata.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of MACS to use. Can be one of :attr:`~MACSCard.SUBSETS`.
        defaults to "audioset_no_audiocaps_v1".
    :param verbose: Verbose level.
        defaults to 0.

    :param hf_cache_dir: Optional override for HuggingFace cache directory path.
        defaults to None.
    :param revision: Optional override for revision commit/name for HuggingFace rapository.
        defaults to None.
    :returns: A dictionnary of lists containing each metadata.
    """

    if subset in _WAVCAPS_OLD_SUBSETS_NAMES:
        new_subset = _WAVCAPS_OLD_SUBSETS_NAMES[subset]
        if verbose >= 0:
            msg = f"Deprecated subset name '{subset}', use '{new_subset}' instead."
            logger.warning(msg)
        subset = new_subset

    root = _get_root(root)
    if subset not in WavCapsCard.SUBSETS:
        msg = f"Invalid argument {subset=}. (expected one of {WavCapsCard.SUBSETS})"
        raise ValueError(msg)

    if subset == "audioset":
        overlapped_ds = "AudioCaps (v1 and v2)"
        overlapped_subsets = ("val", "test")
        recommanded = "audioset_no_audiocaps_v1"
        msg = (
            f"You selected WavCaps subset '{subset}', be careful to not use these data as training when evaluating on {overlapped_ds} {overlapped_subsets} subsets. "
            f"You can use {recommanded} subset for to avoid this bias with {overlapped_ds}."
        )
        logger.warning(msg)

    elif subset == "freesound":
        overlapped_ds = "Clotho"
        overlapped_subsets = (
            "val",
            "eval",
            "dcase_aac_test",
            "dcase_aac_analysis",
            "dcase_t2a_audio",
            "dcase_t2a_captions",
        )
        recommanded = "freesound_no_clotho_v2"
        msg = (
            f"You selected WavCaps subset '{subset}', be careful to not use these data as training when evaluating on {overlapped_ds} {overlapped_subsets} subsets. "
            f"You can use {recommanded} subset for to avoid this bias for Clotho val, eval, dcase_t2a_audio and dcase_t2a_captions subsets. Data could still overlap with Clotho dcase_aac_test and dcase_aac_analysis subsets."
        )
        logger.warning(msg)

    if subset in (
        "audioset_no_audiocaps_v1",
        "freesound_no_clotho_v2",
    ):
        if subset == "audioset_no_audiocaps_v1":
            target_subset = "audioset"
            csv_fname = _WAVCAPS_LINKS["blacklist_audiocaps"]["fname"]

        elif subset == "freesound_no_clotho_v2":
            target_subset = "freesound"
            csv_fname = _WAVCAPS_LINKS["blacklist_clotho_v2"]["fname"]

        else:
            msg = f"INTERNAL ERROR: Invalid argument {subset=}."
            raise ValueError(msg)

        raw_data = _load_wavcaps_dataset_impl(
            root=root,
            subset=target_subset,
            verbose=verbose,
            hf_cache_dir=hf_cache_dir,
            revision=revision,
        )
        wavcaps_ids = raw_data["id"]

        wavcaps_root = _get_wavcaps_root(root, hf_cache_dir, revision)
        csv_fpath = Path(wavcaps_root).joinpath(csv_fname)

        with open(csv_fpath, "r") as file:
            reader = csv.DictReader(file)
            data = list(reader)

        other_ids = [data_i["id"] for data_i in data]
        other_ids = dict.fromkeys(other_ids)

        indexes = [i for i, wc_id in enumerate(wavcaps_ids) if wc_id not in other_ids]

        if verbose >= 1:
            msg = f"Getting {len(indexes)}/{len(wavcaps_ids)} items from '{target_subset}' for subset '{subset}'."
            logger.info(msg)

        raw_data = {
            column: [column_data[index] for index in indexes]
            for column, column_data in raw_data.items()
        }
        return raw_data

    raw_data = _load_wavcaps_dataset_impl(
        root=root,
        subset=subset,
        verbose=verbose,
        hf_cache_dir=hf_cache_dir,
        revision=revision,
    )
    return raw_data


def download_wavcaps_dataset(
    # Common args
    root: Union[str, Path, None] = None,
    subset: WavCapsSubset = WavCapsCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    verify_files: bool = False,
    *,
    # WavCaps-specific args
    clean_archives: bool = False,
    hf_cache_dir: Optional[str] = None,
    repo_id: Optional[str] = None,
    revision: Optional[str] = None,
    zip_path: Union[str, Path, None] = None,
) -> None:
    """Prepare WavCaps data.

    :param root: Dataset root directory.
        defaults to ".".
    :param subset: The subset of MACS to use. Can be one of :attr:`~WavCapsCard.SUBSETS`.
        defaults to "audioset_no_audiocaps_v1".
    :param force: If True, force to download again all files.
        defaults to False.
    :param verbose: Verbose level.
        defaults to 0.
    :param verify_files: If True, check all file already downloaded are valid.
        defaults to False.

    :param clean_archives: If True, remove the compressed archives from disk to save space.
        defaults to True.
    :param hf_cache_dir: Optional override for HuggingFace cache directory path.
        defaults to None.
    :param repo_id: Repository ID on HuggingFace.
        defaults to "cvssp/WavCaps".
    :param revision: Optional override for revision commit/name for HuggingFace rapository.
        defaults to None.
    :param zip_path: Path to zip executable path in shell.
        defaults to "zip".
    """

    if subset in _WAVCAPS_OLD_SUBSETS_NAMES:
        new_subset = _WAVCAPS_OLD_SUBSETS_NAMES[subset]
        if verbose >= 0:
            msg = f"Deprecated subset name '{subset}', use '{new_subset}' instead."
            logger.warning(msg)
        subset = new_subset

    root = _get_root(root)
    zip_path = _get_zip_path(zip_path)

    if subset == "audioset_no_audiocaps_v1":
        _download_blacklist(root, hf_cache_dir, revision, "blacklist_audiocaps")

        return download_wavcaps_dataset(
            root=root,
            subset="audioset",
            revision=revision,
            hf_cache_dir=hf_cache_dir,
            force=force,
            verify_files=verify_files,
            clean_archives=clean_archives,
            zip_path=zip_path,
            verbose=verbose,
        )

    elif subset == "freesound_no_clotho_v2":
        _download_blacklist(root, hf_cache_dir, revision, "blacklist_clotho_v2")

        return download_wavcaps_dataset(
            root=root,
            subset="freesound",
            revision=revision,
            hf_cache_dir=hf_cache_dir,
            force=force,
            verify_files=verify_files,
            clean_archives=clean_archives,
            zip_path=zip_path,
            verbose=verbose,
        )

    if subset not in WavCapsCard.SUBSETS:
        msg = f"Invalid argument {subset=}. (expected one of {WavCapsCard.SUBSETS})"
        raise ValueError(msg)

    # note: verbose=-1 to disable warning triggered when dset is not prepared
    if not force and _is_prepared_wavcaps(
        root, hf_cache_dir, revision, subset, verbose=-1
    ):
        return None

    if hf_cache_dir is None:
        hf_cache_dir = HUGGINGFACE_HUB_CACHE
    if repo_id is None:
        repo_id = WavCapsCard.REPO_ID

    # Download files from huggingface
    ign_sources = [
        source for source in WavCapsCard.SOURCES if not _use_source(source, subset)
    ]
    ign_patterns = [
        pattern
        for source in ign_sources
        for pattern in (f"json_files/{source}/*.json", f"Zip_files/{source}/*")
    ]
    if verbose >= 2:
        logger.debug(f"ign_sources={ign_sources}")
        logger.debug(f"ign_patterns={ign_patterns}")

    pbar_enabled = are_progress_bars_disabled()
    if pbar_enabled and verbose <= 0:
        disable_progress_bars()

    snapshot_dpath = snapshot_download(
        repo_id=repo_id,
        repo_type="dataset",
        revision=revision,
        resume_download=not force,
        local_files_only=False,
        cache_dir=hf_cache_dir,
        allow_patterns=None,
        ignore_patterns=ign_patterns,
    )

    if pbar_enabled and verbose <= 0:
        enable_progress_bars()

    snapshot_abs_dpath = osp.abspath(snapshot_dpath)
    wavcaps_root = _get_wavcaps_root(root, hf_cache_dir, revision)
    if verbose >= 2:
        logger.debug(f"snapshot_dpath={snapshot_dpath}")
        logger.debug(f"snapshot_absdpath={snapshot_abs_dpath}")
        logger.debug(f"wavcaps_dpath={wavcaps_root}")
    del snapshot_dpath

    # Build symlink to hf cache
    if osp.lexists(wavcaps_root):
        if not osp.islink(wavcaps_root):
            raise RuntimeError("WavCaps root exists but it is not a symlink.")
        link_target_abspath = osp.abspath(osp.realpath(wavcaps_root))
        if link_target_abspath != snapshot_abs_dpath:
            logger.error(
                "Target link is not pointing to current snapshot path. It will be automatically replaced."
            )
            os.remove(wavcaps_root)
            os.symlink(snapshot_abs_dpath, wavcaps_root, True)
    else:
        os.symlink(snapshot_abs_dpath, wavcaps_root, True)

    source_and_splitted_lst: List[Tuple[WavCapsSource, bool]] = [
        ("AudioSet_SL", True),
        ("BBC_Sound_Effects", True),
        ("FreeSound", True),
        ("SoundBible", False),
    ]
    source_and_splitted: Dict[WavCapsSource, bool] = {
        source: is_splitted
        for source, is_splitted in source_and_splitted_lst
        if _use_source(source, subset)
    }

    archives_dpath = _get_archives_dpath(root, hf_cache_dir, revision)
    for source, is_splitted in source_and_splitted.items():
        main_zip_fpath = osp.join(
            archives_dpath, _WAVCAPS_ARCHIVE_DNAMES[source], f"{source}.zip"
        )

        if is_splitted:
            merged_zip_fpath = osp.join(
                archives_dpath, _WAVCAPS_ARCHIVE_DNAMES[source], f"{source}_merged.zip"
            )
        else:
            merged_zip_fpath = main_zip_fpath

        if is_splitted and not osp.isfile(merged_zip_fpath):
            cmd = [
                zip_path,
                "-FF",
                main_zip_fpath,
                "--out",
                merged_zip_fpath,
            ]
            if verbose >= 2:
                logger.debug(f"Merging ZIP files for {source}...")
                logger.debug(f"Using command: {' '.join(cmd)}")

            if verbose >= 2:
                stdout = None
                stderr = None
            else:
                stdout = subprocess.DEVNULL
                stderr = subprocess.DEVNULL

            subprocess.check_call(cmd, stdout=stdout, stderr=stderr)

        audio_subset_dpath = _get_audio_subset_dpath(
            root, hf_cache_dir, revision, source
        )
        os.makedirs(audio_subset_dpath, exist_ok=True)

        with zipfile.ZipFile(merged_zip_fpath, "r") as file:
            flac_subnames = [name for name in file.namelist() if name.endswith(".flac")]
            assert len(flac_subnames) > 0
            assert all(
                osp.dirname(name) == osp.dirname(flac_subnames[0])
                for name in flac_subnames
            )

            src_root = osp.join(audio_subset_dpath, osp.dirname(flac_subnames[0]))
            src_fnames_found = (
                dict.fromkeys(name for name in os.listdir(src_root))
                if osp.isdir(src_root)
                else {}
            )
            tgt_fnames_found = dict.fromkeys(
                name for name in os.listdir(audio_subset_dpath)
            )

            missing_subnames = [
                subname
                for subname in flac_subnames
                if osp.basename(subname) not in src_fnames_found
                and osp.basename(subname) not in tgt_fnames_found
            ]
            if verbose >= 2:
                logger.debug(
                    f"Extracting {len(missing_subnames)}/{len(flac_subnames)} audio files from {merged_zip_fpath}..."
                )
            file.extractall(audio_subset_dpath, missing_subnames)
            if verbose >= 2:
                logger.debug("Extraction done.")

        src_fnames_found = (
            dict.fromkeys(name for name in os.listdir(src_root))
            if osp.isdir(src_root)
            else {}
        )
        src_fpaths_to_move = [
            osp.join(audio_subset_dpath, subname)
            for subname in flac_subnames
            if osp.basename(subname) in src_fnames_found
        ]
        if verbose >= 2:
            logger.debug(f"Moving {len(src_fpaths_to_move)} files...")
        for src_fpath in tqdm.tqdm(src_fpaths_to_move):
            tgt_fpath = osp.join(audio_subset_dpath, osp.basename(src_fpath))
            os.rename(src_fpath, tgt_fpath)
        if verbose >= 2:
            logger.debug("Move done.")

        if verify_files:
            tgt_fnames_expected = [osp.basename(subname) for subname in flac_subnames]
            tgt_fnames_found = dict.fromkeys(
                fname for fname in os.listdir(audio_subset_dpath)
            )
            if verbose >= 2:
                logger.debug(f"Checking {len(tgt_fnames_expected)} files...")
            tgt_fnames_invalids = [
                fname for fname in tgt_fnames_expected if fname not in tgt_fnames_found
            ]
            if len(tgt_fnames_invalids) > 0:
                raise FileNotFoundError(
                    f"Found {len(tgt_fnames_invalids)}/{len(tgt_fnames_expected)} invalid files."
                )

        safe_rmdir(audio_subset_dpath, rm_root=False, error_on_non_empty_dir=True)

    if clean_archives:
        used_sources = source_and_splitted.keys()
        for source in used_sources:
            archive_source_dpath = osp.join(
                archives_dpath, _WAVCAPS_ARCHIVE_DNAMES[source]
            )
            archives_names = os.listdir(archive_source_dpath)
            for name in archives_names:
                if not name.endswith(".zip") and ".z" not in name:
                    continue
                fpath = osp.join(archive_source_dpath, name)
                if verbose >= 1:
                    logger.info(f"Removing archive file {name} for {source=}...")
                os.remove(fpath)


def download_wavcaps_datasets(
    # Common args
    root: Union[str, Path, None] = None,
    subsets: Union[WavCapsSubset, Iterable[WavCapsSubset]] = WavCapsCard.DEFAULT_SUBSET,
    force: bool = False,
    verbose: int = 0,
    *,
    # WavCaps-specific args
    clean_archives: bool = False,
    hf_cache_dir: Optional[str] = None,
    repo_id: Optional[str] = None,
    revision: Optional[str] = None,
    verify_files: bool = False,
    zip_path: Union[str, Path, None] = None,
) -> None:
    """Function helper to download a list of subsets. See :func:`~aac_datasets.datasets.functional.wavcaps.download_wavcaps_dataset` for details."""
    if isinstance(subsets, str):
        subsets = [subsets]
    else:
        subsets = list(subsets)

    kwargs: Dict[str, Any] = dict(
        root=root,
        force=force,
        verbose=verbose,
        clean_archives=clean_archives,
        hf_cache_dir=hf_cache_dir,
        repo_id=repo_id,
        revision=revision,
        verify_files=verify_files,
        zip_path=zip_path,
    )
    for subset in subsets:
        download_wavcaps_dataset(
            subset=subset,
            **kwargs,
        )


def _load_wavcaps_dataset_impl(
    # Common args
    root: str,
    subset: WavCapsSubset,
    verbose: int,
    # WavCaps-specific args
    hf_cache_dir: Optional[str],
    revision: Optional[str],
) -> Dict[str, List[Any]]:
    if not _is_prepared_wavcaps(root, hf_cache_dir, revision, subset, verbose):
        msg = f"{WavCapsCard.PRETTY_NAME} is not prepared in {root=}. Please use download=True to install it in root."
        raise RuntimeError(msg)

    json_dpath = _get_json_dpath(root, hf_cache_dir, revision)
    json_paths: List[Tuple[WavCapsSource, str]] = [
        ("AudioSet_SL", osp.join(json_dpath, "AudioSet_SL", "as_final.json")),
        (
            "BBC_Sound_Effects",
            osp.join(json_dpath, "BBC_Sound_Effects", "bbc_final.json"),
        ),
        ("FreeSound", osp.join(json_dpath, "FreeSound", "fsd_final.json")),
        ("SoundBible", osp.join(json_dpath, "SoundBible", "sb_final.json")),
    ]
    json_paths = [
        (source, json_path)
        for source, json_path in json_paths
        if _use_source(source, subset)
    ]

    raw_data = {k: [] for k in _WAVCAPS_RAW_COLUMNS + ("source", "fname")}
    for source, json_path in json_paths:
        if verbose >= 2:
            logger.debug(f"Loading metadata in JSON '{json_path}'...")
        json_data, size = _load_json(json_path)

        sources = [source] * size
        json_data.pop("audio", None)

        if source == "AudioSet_SL":
            ids = json_data["id"]
            fnames = [id_.replace(".wav", ".flac") for id_ in ids]
            raw_data["fname"] += fnames

        elif source == "BBC_Sound_Effects":
            ids = json_data["id"]
            fnames = [f"{id_}.flac" for id_ in ids]
            raw_data["fname"] += fnames

        elif source == "FreeSound":
            ids = json_data["id"]
            fnames = [f"{id_}.flac" for id_ in ids]
            raw_data["fname"] += fnames

        elif source == "SoundBible":
            ids = json_data["id"]
            fnames = [f"{id_}.flac" for id_ in ids]
            raw_data["fname"] += fnames

        else:
            msg = f"Invalid source={source} in {json_path=}. (expected one of {WavCapsCard.SOURCES})"
            raise RuntimeError(msg)

        for k in _WAVCAPS_RAW_COLUMNS:
            if k in json_data:
                raw_data[k] += json_data[k]
            elif k in _DEFAULT_VALUES:
                default_val = _DEFAULT_VALUES[k]
                default_values = [default_val] * size
                raw_data[k] += default_values
            elif k in ("audio", "file_name"):
                pass
            else:
                raise RuntimeError(f"Invalid column name {k}. (with {source=})")

        raw_data["source"] += sources

    raw_data.pop("audio")
    raw_data.pop("file_name")
    captions = raw_data.pop("caption")

    # Convert str -> List[str] for captions to match other datasets captions type
    raw_data["captions"] = [[caption] for caption in captions]

    # Force floating-point precision for duration
    raw_data["duration"] = list(map(float, raw_data["duration"]))
    return raw_data


def _get_wavcaps_root(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
) -> str:
    return osp.join(root, "WavCaps")


def _get_json_dpath(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
) -> str:
    return osp.join(_get_wavcaps_root(root, hf_cache_dir, revision), "json_files")


def _get_archives_dpath(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
) -> str:
    return osp.join(_get_wavcaps_root(root, hf_cache_dir, revision), "Zip_files")


def _get_audio_dpath(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
) -> str:
    return osp.join(_get_wavcaps_root(root, hf_cache_dir, revision), "Audio")


def _get_audio_subset_dpath(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
    source: WavCapsSource,
) -> str:
    return osp.join(
        _get_audio_dpath(root, hf_cache_dir, revision),
        _WAVCAPS_AUDIO_DNAMES[source],
    )


def _is_prepared_wavcaps(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
    subset: WavCapsSubset,
    verbose: int,
) -> bool:
    sources: List[WavCapsSource] = [
        source for source in WavCapsCard.SOURCES if _use_source(source, subset)
    ]
    for source in sources:
        audio_subset_dpath = _get_audio_subset_dpath(
            root, hf_cache_dir, revision, source
        )
        if not osp.isdir(audio_subset_dpath):
            if verbose >= 0:
                msg = f"Cannot find directory {audio_subset_dpath=}."
                logger.error(msg)
            return False

        audio_fnames = os.listdir(audio_subset_dpath)
        expected_size = WavCapsCard.EXPECTED_SIZES[source]
        if expected_size != len(audio_fnames):
            if verbose >= 0:
                msg = f"Invalid number of files for {source=}. (expected {expected_size} but found {len(audio_fnames)} files)"
                logger.error(msg)
            return False
    return True


def _use_source(source: WavCapsSource, subset: WavCapsSubset) -> bool:
    return any(
        (
            source == "AudioSet_SL"
            and subset in ("audioset", "audioset_no_audiocaps_v1"),
            source == "BBC_Sound_Effects" and subset in ("bbc",),
            source == "FreeSound" and subset in ("freesound", "freesound_no_clotho_v2"),
            source == "SoundBible" and subset in ("soundbible",),
        )
    )


def _load_json(fpath: str) -> Tuple[Dict[str, Any], int]:
    with open(fpath, "r") as file:
        data = json.load(file)
    data = data["data"]
    size = len(data)
    data = pw.list_dict_to_dict_list(data, key_mode="same")
    return data, size


def _download_blacklist(
    root: str,
    hf_cache_dir: Optional[str],
    revision: Optional[str],
    name: str,
    verbose: int = 0,
) -> None:
    info = _WAVCAPS_LINKS[name]
    fname = info["fname"]
    url = info["url"]
    wavcaps_root = _get_wavcaps_root(root, hf_cache_dir, revision)
    fpath = Path(wavcaps_root).joinpath(fname)
    download_file(url, fpath, verbose=verbose)


class _WavCapsRawItem(TypedDict):
    # Common values
    caption: str
    duration: float
    id: str
    # Source Specific values
    audio: Optional[str]
    author: Optional[str]
    description: Optional[str]
    download_link: Optional[str]
    file_name: Optional[str]
    href: Optional[str]
    tags: Optional[List[str]]


_DEFAULT_VALUES = {
    "author": "",
    "description": "",
    "download_link": "",
    "href": "",
    "tags": [],
}

_WAVCAPS_RAW_COLUMNS = tuple(
    _WavCapsRawItem.__required_keys__ | _WavCapsRawItem.__optional_keys__  # type: ignore
)

_WAVCAPS_AUDIO_DNAMES: Dict[WavCapsSource, str] = {
    # Source name to audio directory name
    "AudioSet_SL": "AudioSet_SL",
    "BBC_Sound_Effects": "BBC_Sound_Effects",
    "FreeSound": "FreeSound",
    "SoundBible": "SoundBible",
}

_WAVCAPS_ARCHIVE_DNAMES: Dict[WavCapsSource, str] = {
    # Source name to audio directory name
    "AudioSet_SL": "AudioSet_SL",
    "BBC_Sound_Effects": "BBC_Sound_Effects",
    "FreeSound": "FreeSound",
    "SoundBible": "SoundBible",
}

_WAVCAPS_LINKS: Dict[str, LinkInfo] = {
    "blacklist_audiocaps": {
        "url": "https://raw.githubusercontent.com/Labbeti/aac-datasets/main/data/wavcaps/blacklist_audiocaps.full.csv",
        "fname": "blacklist_audiocaps.full.csv",
    },
    "blacklist_clotho": {
        "url": "https://raw.githubusercontent.com/Labbeti/aac-datasets/main/data/wavcaps/blacklist_clotho.full.csv",
        "fname": "blacklist_clotho.full.csv",
    },
    "blacklist_clotho_v2": {
        "url": "https://raw.githubusercontent.com/Labbeti/aac-datasets/main/data/wavcaps/blacklist_clotho.full.v2.csv",
        "fname": "blacklist_clotho.full.v2.csv",
    },
}

_WAVCAPS_OLD_SUBSETS_NAMES: Dict[str, WavCapsSubset] = {
    "fsd": "freesound",
    "as": "audioset",
    "fsd_nocl": "freesound_no_clotho_v2",
    "as_noac": "audioset_no_audiocaps_v1",
    "sb": "soundbible",
    "audioset_no_audiocaps": "audioset_no_audiocaps_v1",
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os.path as osp
from pathlib import Path
from typing import Any, Callable, ClassVar, List, Optional, Union

from torch import Tensor
from typing_extensions import TypedDict

from aac_datasets.datasets.base import AACDataset
from aac_datasets.datasets.functional.wavcaps import (
    WavCapsCard,
    WavCapsSource,
    WavCapsSubset,
    _get_audio_subset_dpath,
    download_wavcaps_dataset,
    load_wavcaps_dataset,
)
from aac_datasets.utils.globals import _get_root, _get_zip_path

logger = logging.getLogger(__name__)


class WavCapsItem(TypedDict):
    # Common attributes
    audio: Tensor
    captions: List[str]
    dataset: str
    fname: str
    index: int
    subset: WavCapsSubset
    sr: int
    duration: float
    # WavCaps-specific attributes
    author: Optional[str]  # FSD and SB
    description: Optional[str]  # BBC, FSD and SB only
    duration: float
    download_link: Optional[str]  # BBC, FSD and SB only
    href: Optional[str]  # FSD and SB only
    id: str
    source: WavCapsSource
    tags: List[str]  # FSD only


class WavCaps(AACDataset[WavCapsItem]):
    r"""Unofficial WavCaps PyTorch dataset.

    WavCaps Paper : https://arxiv.org/pdf/2303.17395.pdf
    HuggingFace source : https://huggingface.co/datasets/cvssp/WavCaps

    This dataset contains 4 training subsets, extracted from different sources:
    - BBC Sound Effects "bbc"
    - SoundBible "soundbible"
    - AudioSet strongly labeled without AudioCaps V1 val and test subsets "audioset_no_audiocaps_v1"
    - FreeSound without Clotho dev, val, eval and test subsets "freesound_no_clotho_v2"

    Other subsets exists but they does not comply DCASE Challenge rules:
    - AudioSet strongly labeled "audioset"
    - FreeSound "freesound"

    .. warning::
        WavCaps download is experimental ; it requires a lot of disk space and can take very long time to download and extract, so you might expect errors.

    .. code-block:: text
        :caption:  Dataset folder tree

        {root}
        └── WavCaps
            ├── Audio
            │   ├── AudioSet_SL
            │   │    └── (108317 flac files, ~64GB)
            │   ├── BBC_Sound_Effects
            │   │    └── (31201 flac files, ~142GB)
            │   ├── FreeSound
            │   │    └── (262300 flac files, ~1.4TB)
            │   └── SoundBible
            │        └── (1232 flac files, ~884MB)
            ├── Zip_files
            │   ├── AudioSet_SL
            │   │    └── (8 zip files, ~76GB)
            │   ├── BBC_Sound_Effects
            │   │    └── (26 zip files, ~562GB)
            │   ├── FreeSound
            │   │    └── (123 zip? files, ~1.4TB)
            │   └── SoundBible
            │        └── (1 zip? files, ~624GB)
            ├── json_files
            │    ├── AudioSet_SL
            │    │    └── as_final.json
            │    ├── BBC_Sound_Effects
            │    │    └── bbc_final.json
            │    ├── FreeSound
            │    │    ├── fsd_final_2s.json
            │    │    └── fsd_final.json
            │    ├── SoundBible
            │    │    └── sb_final.json
            │    └── blacklist
            │         ├── blacklist_exclude_all_ac.json
            │         ├── blacklist_exclude_test_ac.json
            │         └── blacklist_exclude_ubs8k_esc50_vggsound.json
            ├── .gitattributes
            └── README.md
    """

    # Common globals
    CARD: ClassVar[WavCapsCard] = WavCapsCard()

    def __init__(
        self,
        # Common args
        root: Union[str, Path, None] = None,
        subset: WavCapsSubset = WavCapsCard.DEFAULT_SUBSET,
        download: bool = False,
        transform: Optional[Callable[[WavCapsItem], Any]] = None,
        verbose: int = 0,
        force_download: bool = False,
        verify_files: bool = False,
        *,
        # WavCaps-specific args
        clean_archives: bool = False,
        hf_cache_dir: Optional[str] = None,
        repo_id: Optional[str] = None,
        revision: Optional[str] = WavCapsCard.DEFAULT_REVISION,
        zip_path: Union[str, Path, None] = None,
    ) -> None:
        """
        :param root: The parent of the dataset root directory.
            The data will be stored in the 'MACS' subdirectory.
            defaults to ".".
        :param subset: The subset of the dataset. Can be one of :attr:`~WavCapsCard.SUBSETS`.
            defaults to "audioset".
        :param download: Download the dataset if download=True and if the dataset is not already downloaded.
            defaults to False.
        :param transform: The transform to apply to the global dict item. This transform is applied only in getitem method when argument is an integer.
            defaults to None.
        :param verbose: Verbose level. Can be 0 or 1.
            defaults to 0.
        :param force_download: If True, force to re-download file even if they exists on disk.
            defaults to False.
        :param verify_files: If True, check hash value when possible.
            defaults to False.

        :param clean_archives: If True, remove the compressed archives from disk to save space.
            defaults to False.
        :param hf_cache_dir: HuggingFace cache directory. If None, use the global value :variable:`~huggingface_hub.constants.HUGGINGFACE_HUB_CACHE`.
            defaults to None.
        :param repo_id: Repository ID on HuggingFace.
            defaults to "cvssp/WavCaps".
        :param revision: The HuggingFace revision tag.
            defaults to :attr:`~WavCapsCard.DEFAULT_REVISION`.
        :param zip_path: Path to zip executable path in shell.
            defaults to "zip".
        """
        if subset not in WavCapsCard.SUBSETS:
            msg = f"Invalid argument {subset=} for {WavCapsCard.PRETTY_NAME}. (expected one of {WavCapsCard.SUBSETS})"
            raise ValueError(msg)

        root = _get_root(root)
        zip_path = _get_zip_path(zip_path)

        if download:
            download_wavcaps_dataset(
                root=root,
                subset=subset,
                force=force_download,
                verbose=verbose,
                clean_archives=clean_archives,
                hf_cache_dir=hf_cache_dir,
                repo_id=repo_id,
                revision=revision,
                verify_files=verify_files,
                zip_path=zip_path,
            )

        raw_data = load_wavcaps_dataset(
            root=root,
            subset=subset,
            verbose=verbose,
            hf_cache_dir=hf_cache_dir,
            revision=revision,
        )

        size = len(next(iter(raw_data.values())))
        raw_data["dataset"] = [WavCapsCard.NAME] * size
        raw_data["subset"] = [subset] * size
        raw_data["fpath"] = [
            osp.join(
                _get_audio_subset_dpath(
                    root, hf_cache_dir, revision, raw_data["source"][i]
                ),
                fname,
            )
            for i, fname in enumerate(raw_data["fname"])
        ]
        raw_data["index"] = list(range(size))

        column_names = list(WavCapsItem.__required_keys__) + list(  # type: ignore
            WavCapsItem.__optional_keys__  # type: ignore
        )

        super().__init__(
            raw_data=raw_data,
            transform=transform,
            column_names=column_names,
            flat_captions=False,
            sr=WavCapsCard.SAMPLE_RATE,
            verbose=verbose,
        )
        self._root = root
        self._subset: WavCapsSubset = subset
        self._download = download
        self._hf_cache_dir = hf_cache_dir
        self._revision = revision

        self.add_online_columns(
            {
                "audio": WavCaps._load_audio,
                "audio_metadata": WavCaps._load_audio_metadata,
                "duration": WavCaps._load_duration,
                "num_channels": WavCaps._load_num_channels,
                "num_frames": WavCaps._load_num_frames,
                "sr": WavCaps._load_sr,
            }
        )

    # Properties
    @property
    def download(self) -> bool:
        return self._download

    @property
    def root(self) -> str:
        return self._root

    @property
    def sr(self) -> int:
        return self._sr  # type: ignore

    @property
    def subset(self) -> WavCapsSubset:
        return self._subset
